include("helpers.jl")

##############################################################
#
# Implements some fast matrix completion algorithms
#
###############################################################



##########################  Basic AM ###############################################

# Intializes AM as in [JNS'13]; takes an SVD of a subsampled matrix and zeros out the large entries.
# A is a sparse matrix
# k is the target rank
# L is the number of iterations for power method to approximate SVD
# mu is a bound on the coherence of the original (dense) matrix 
function svdStart(A, k, L, mu)
	n = size(A,1)
	B,sigmas = svdPowerMethod(A, k, L)
	thresh = 2*mu*sqrt(k/n)
	trunc( B, thresh, true )
	return B
end

# return Y in R^{kxn} which minimimizes || (XY)_Omega - A ||_F^2
# A is an nxn sparse matrix
# X is an nxk dense matrix
function amIterate( A, X ,debug=false)
	n = size(A,1)
	D = colLists(A)  # get a dict of the columns. 
	Y = zeros(size(X,2),size(X,1)) 
	for j = 1:n

		if haskey(D,j)
			# make the projector
			Pj = spzeros(n,n)
			Aj = D[j]  # jth column of A
			for a in findn(Aj)[1]  # nonzero entries in jth column
				Pj[a,a] = 1.0
			end
			B = Pj * X
			# the j'th column of Y is y = argmin || By = Aj ||_2
			#y = pinv( B ) * (Aj)
			## doing it this way seems to be faster:
			y = 0
			BTB = transpose(B) * B
			try
				y = inv( BTB ) * ( transpose(B) * Aj )
			catch
				throw(DivideError()) 
				return 
			end
			Y[:,j] = y
		end
	end
	return Y  # Y is kxn,
end		

# Alternating minimization, randomly initialized
# A is a sparse matrix, samples of a rank-k matrix
# miter is the number of observations PER ITERATION
# k is the target rank
# mu is a bound on the coherence of the original matrix A
# if the debug flag is true, you can also pass in a correctX (the right answer) and a Tracker2 data structure to collect data.
#
# Returns X,Y, so that XY^T is an estimate of the original matrix.
function basicAM( A, miter, k; mu=1, debug=false, correctX=-1, tracker=-1 ,correctA=-1,domedian=-1)
	n = size(A,1)
	# chop up the samples in A
	m = nnz(A) 
	L = convert( Int, floor( m/miter ) )
	pieces = chopUp(A, [miter for i=1:L]  )

	# random starting point.
	X = randomOrthonormal(n,k)

	# do AM.
	X,Y,samples = AM( pieces, k, X, debug=debug,correctX=correctX,correctA=correctA, tracker=tracker,domedian=domedian)
	return X, Y , samples
end

# Alternating minimization, initialized as in [JNS'13]
# A is a sparse matrix (the samples)
# Lstart is the number of iterations for the power method in svdStart
# mstart is the number of observations for svdStart
# miter is the number of observations to use per iteration of AM
# k is the target rank
# mu is a bound on the coherence of the original matrix A
function initializedAM( A, mstart, miter, k; Lstart=100, mu=1,debug=false, correctX=-1,tracker=-1,correctA=-1,domedian=-1 )
	n = size(A,1)

	# chop up the samples in A
	m = nnz(A) 
	L = convert( Int,floor( (m - mstart)/miter ) )
	pieces = chopUp(A, cat(1, [mstart], [miter for i=1:L] ) )

	# initialize, and update the tracker appropriately.
	X = svdStart( pieces[1], k, Lstart, mu )
	initsamples = nnz(pieces[1])
	if debug && correctX!=-1 && tracker!=-1
		updateTracker( tracker, correctX, X, "Deflate" )
	end

	# do AM.
	X,Y,samples = AM( pieces[2:end], k, X, debug=debug,correctX=correctX,correctA=correctA, tracker=tracker,domedian=domedian)
	return X, Y, samples + initsamples
end

# Alternating minimization.
# Apieces is an array of sparse matrices. 
# k is the target rank
# Xstart is the starting subspace
# debug is a flag for debugging
# correctX is the (optional) correct answer (for debugging), same for correctA
# tracker is a Tracker2 data structure, for collecting data internal to the alg.
# if the median is an int!=-1, then break the observations into that many pieces and do the median stuff from [Hardt'14]
function AM( Apieces::Array{SparseMatrixCSC{Any,Int64},1}, k::Int, Xstart::Array{Float64} ; debug=false, correctX=-1, correctA=-1,tracker=-1, domedian=-1)
	L = length(Apieces)
	n = size(Apieces[1] , 1)

	X = Xstart
	Xprev = Xstart
	Y = zeros( Float64, k,n )	
	nsamples = 0
	for l=1:length(Apieces)
		B = Apieces[l]
		Xprev = X
		try
			samples = nnz(B) 
			nsamples += samples
			if debug
				write("\t\t AM iteration ")
				write(string(l))
				write(": ")
				write( string( samples))
				write( " samples")
			end
			if domedian != -1
				if debug
					write("\n")
				end
				sizemedian = convert(Int, floor( nnz(B) / domedian ) )
				subPieces = chopUp( B, [ sizemedian for i=1:domedian ] )
				subYs = Array( Array{Float64} , domedian )
				for i=1:length(subPieces)
					subB = subPieces[i]
					if debug
						println("\t\t\t Median iteration ", i, ": ", nnz(subB), " samples")
					end
					subY = amIterate( subB, X, debug )
					subYs[i] = subY
				end
				Y= entrywiseMedian( subYs ) 
			else
				Y = amIterate( B, X ,debug)
			end
		catch anError
			if isa(anError, DivideError)
				if true #debug  
					println("\t\t ...Skipping this iteration...couldn't invert....")
				end
				continue
			else
				rethrow(anError)
			end
		end
		X,R = qr(transpose(Y))
		if debug  
			if correctA != -1
				write("\t ||XY - A||_F=")
				write(string( normfro(Xprev*Y - A)))
			end
			if correctX != -1
				write ("\t pangles:")
				pangles = svd( X - correctX * (transpose(correctX) * X) )[2]
				for pangle in pangles
					write( string(round(pangle,2)) )
					write(" ")
				end
				if tracker!= -1
					updateTracker( tracker, correctX, X, "AM") # tracks the principal angles.
				end
			end
			write("\n")
		end
	end
	return Xprev, Y, nsamples # X*Y is an approximation of A
end

############################### Deflation-based AM ###################################


# Deflation-based algorithm 
#
# This has no smoothing, and no medians. (Yet?)
#
# A is a sparse matrix
# mDeflate is the number of observations for each deflation step
# mAM is the number of observations in each AM iteration.
# mustar is an upper bound on the coherence of A; it's 1 by default.
# mu0 is larger than C( mustar k^(3/2) + log(n) ).  Right now we are not using it (smoothing not implemented).
# epslevel is the accuracy parameter; by default it is tinier than we will run into. 
# T, if set, is the number of times to run altmin.  By default this is k.
#     Notice that the TOTAL number of measurements is bounded above by m*(L+1)*T + m; 
#     first, use m measurements to initialize; then, for each run of AM, do L trials and 1 SVD, each with m observations.  
# debug is a flag for debug mode; in debug mode, you may optionally specify the "correct" answers 
#     and give a Tracker data structure to report the error and how it evolves.
# Lsvd is the number of iterations to use in the power method to approximate SVD; it's 100 by default.
function deflationAM( A, mDeflate, mAM ,k; mustar=1, mu0=1, epslevel=.00000000001 ,T=k, debug=false, correctX=-1, correctSigmas=-1, tracker=-1,Lsvd=100)
	n = size(A,1)
	normalizationFactorSVD = n^2/mDeflate
	nsamples = 0

	if debug  # the debugging stuff really wants to know the correct answers...
		if correctX == -1 || correctSigmas==-1
			println("Really, just pass in the correct X please.")
		end
	end

	# use mDeflate for the first step, and then (mDeflate + L(mAM))T for the rest.
	L = convert( Int, floor( ((nnz(A) - mDeflate )/T - mDeflate )/mAM ) )  # how many rounds can we afford?
	if L <= 0
		error("mDeflate and mIter are too big to do even a single iteration of AM")
	end
	pieces = chopUp( A, cat(1, [mDeflate for t=1:T+1 ], [mAM for t=1:L*T] ) )
	deflatePieces = pieces[1:T+1] # we will use the last piece for the first init, and the rest in order.
	AMPieces = pieces[T+2:end]

	if debug
		println("n=", n, " mAM=", mAM, " mDeflate=",mDeflate, " L=", L)
	end
	
	# initialize parameters
	U0, sigmas0 = svdPowerMethod( convert( SparseMatrixCSC{Float64,Int}, deflatePieces[T+1]), 1, Lsvd ) # there is no "norm" for sparse matrices in julia?
	s0 = sigmas0[1] * n^2/mDeflate  # s0 is an estimate of ||A|| 
	s = s0
	nsamples += nnz( deflatePieces[T+1] )
	if debug
		println("Approximation of norm: ", s0, ".  Actual norm: ", correctSigmas[1] )
	end
	X = zeros( Float64, n, k )
	Y = zeros( Float64, k, n )
	r = 0

	# start the AM epochs
	for t=1:T
		deflateA = convert( SparseMatrixCSC{Float64,Int}, deflatePieces[t] ) 
		#amParts = [ convert( SparseMatrixCSC{Float64,Int64}, SP ) for SP in AMPieces[(t-1)L + 1: t*L] ]
		amParts = AMPieces[(t-1)L + 1: t*L] 

		if debug
			println("iteration ",string(t), ": ")
			#println("\t frob error: ", normfro( A - X*Y ) )
			println("\t nsamples so far: ", nsamples )
			println("\t sin angle: ", sinTheta(X, correctX[:, 1:size(X,2) ] ) )
			println("\t r=",r)
		end

		# Are we there yet?
		if r >= k
			if debug
				println("done because r>=k: r=", r, " and k=",k)
			end
			return X, Y, nsamples
		end

		# Tmat = sample(A - XY) is a sample of the part of A we've not gotten already.
		xs,ys = findn( deflateA )
		subsampleXY = subsampleFactor(X, Y, xs,ys )
		Tmat = deflateA - subsampleXY
		nsamples += nnz(deflateA)

		## step 1: estimate the gaps by finding the SVD of the truncated residual
		tau = mustar*2*k*s/n  # truncation threshold; unnormalized, since we didn't normalize T 
		trunc( Tmat, tau )
		Uapprox, sigmaApprox = svdPowerMethod( Tmat, k-r, Lsvd )
		if debug
			println("\t ...number samples for SVD approx:", nnz(Tmat))
			write("\t ...approximate remaining spectrum: ")
			for x in sigmaApprox
				write( string(x * normalizationFactorSVD ) )
				write(" ")
			end
			write("\n")
		end
		# are we there yet?
		if sigmaApprox[1]*normalizationFactorSVD < epslevel*s0  
			if debug
				println("done because estimated sigma1 is small: ", sigmaApprox[1]*normalizationFactorSVD, "<", epslevel*s0 )
			end
			return X, Y, nsamples
		end
		d = -1
		for i in 1:k-r-1
			#if sigmaApprox[i+1] <= (1.0 - 1.0/(4*k))*sigmaApprox[i]
			if sigmaApprox[i+1] <= (1.0 - 1.0/(2*k)) * sigmaApprox[i]
				d = i
				break
			end
		end
		if d == -1
			d = k - r
		end
		if debug
			println("\t ...identified a gap at d=", d)
		end
		# d is the space until the next gap; update parameters appropriately
		r = r + d
		s = sigmaApprox[d]* normalizationFactorSVD  # s is an approximation of the last big singular value

		## step 2: use the first d columns of Uapprox to get a good starting point for AM
		Qapprox = Uapprox[:,1:d]
		trunc( Qapprox, sqrt( mustar * k / n ) )
		W = hcat(X[:,1:r-d], Qapprox)
		Wbar, Rtmp = qr(W)
		if debug
			println("\t ...new starting point has sin error:", sinTheta( Wbar, correctX[:, 1:r] ) ) 
			if tracker!= -1
				updateTracker( tracker, correctX[:,1:r], Wbar , "Deflate")
			end
		end
		# TODO right now we are not using mut, because smoothQR is not used.
		#  mut = ( sqrt( mu0 ) + (t-1)*sqrt(mustar * k ) )^2

		## step 3: use the starting point to run AM and get a new estimates X,Y
		X,Y,samples = AM( amParts, r, Wbar, debug=debug, correctX=correctX,tracker=tracker)
		nsamples += samples
	end
	if debug
		println("Done because T ran out.")
		println("Total number of samples:", nsamples)
	end
	return X,Y, nsamples
end

