## Does "Hazan's algorithm," as in [Jaggi-Sulovsky], for symmetric matrices.
#
# A is a sparsification of a symmetric trace-1 matrix
# eps is an accuracy parameter.
#
# Returns a trace-1 rank-1/eps matrix which should approximate the original.
function FW( A, eps; debug=false, correctQ=-1, correctS=-1, tracker=tracker)
	nsamples = nnz(A)
	Ax,Ay,vals = findnz(A)
	# because I only implemented the symmetric version I need for the sparse version to be symmetric.
	A = map( max, A, transpose(A))
	Ax,Ay = findn(A)
	##### SANITY CHECK
	if correctQ!=-1 && correctS !=-1
		row = correctQ[ Ax[1], : ]
		col = transpose(correctQ[ Ay[1], : ])
		if abs(A[Ax[1],Ay[1]] - (row*diagm(correctS)*col)[1,1])  > .0000001
			error(string( "Something has gone wrong (probably with normalization) in FW"))
		end
	end
	n = size(A,1)
	# initialize
	v = randn(n)
	v = v /norm(v) 
	#Z = v * transpose(v) # random rank-1, trace-1 matrix to start -- don't actually store it.
	steps = convert(Int, round(1.0/eps) )
	coeffs = Array( Float64, steps )
	vecs = Array( Array{Float64}{1}, steps )
	vecs[1] = v
	coeffs[1] = 1.0
	# do FW to search around the simplex
	for j=1:steps-1
		grad = copy(A)
		for i=1:nnz(A)
			grad[Ax[i],Ay[i]] -= getEntry( Ax[i], Ay[i], coeffs, vecs,j ) # Z[Ax[i], Ay[i]]
		end
		# now grad = -gradient( f( Z ) ) , where f = 1/2( fronorm( A_Omega - Z_Omega )^2 )
		evals, evecs, nconv,niter,nmult,resid = eigs( grad, tol=1.0/j^2 )
		evals = [ convert( Float64, eigval ) for eigval in evals ] # they are real since grad is symmetric, just round them.
		# find the largest eigenvalue (might be negative, or zero).
		themax, indexmax = findmax( evals )
		if size(evecs)[2] < n && themax < 0
			# then the biggest eigenvalue is 0, and eigs didn't return the corresponding eigenvector (in Ker(grad))
			M,R = qr( evecs,  false ) # extend the eigenvectors to a real
			v = M[:,n]
		else
			v = evecs[:,indexmax]
		end
		#Delta = (1/j)*( v*transpose(v) - Z )  # update Z
		#Z += Delta
		coeffs = coeffs * (1 - 1/(j+1))
		coeffs[j+1] = 1/(j+1)
		vecs[j+1] = v

		### for debugging: make sure that we actually converge on the samples.
		if debug
			tmperr = 0 
			for i=1:nnz(A)
				tmperr += ( A[Ax[i],Ay[i]] - getEntry( Ax[i], Ay[i], coeffs, vecs,j+1 ) )^2 
			end
			tmperr = convert( Float64, sqrt(tmperr) )
			if correctQ ==-1
				println(j,"\t",tmperr)
			end
			if correctQ != -1
				#fullerr = normfro(getMat( coeffs, vecs, j+1 ) - correctA)
				V = hcat( vecs[1:j+1]... )  # this is a tall skinny matrix
				D = diagm( coeffs[1:j+1] ) 
				fullerr = factorFrob( V, D*transpose(V), correctQ, diagm(correctS)*transpose(correctQ) )  # get frobenius error 
				updateTracker( tracker, [tmperr, fullerr], "" )
				println(j,"\t",tmperr, "\t", fullerr)
			end
		end
	end
	return coeffs, vecs, nsamples
end
