using(Distributions)

function makeRandomMask( n::Int, m::Int )
	if m < n^2 /2
		return makeRandomMaskB(n,m)
	else
		return makeRandomMaskA(n,m)
	end
end
# Give some random coordinates suitable for a sparse matrix constructor 
function makeRandomMaskA( n::Int, m::Int )
	p = m./n^2
	S = Set{Array{Int64,1}}()
	for i=1:n
		for j=1:n
			if rand() < p
				push!(S, [i,j])
			end
		end
	end
	return [ x[1] for x in S ], [x[2] for x in S]
end

## choose the number of guys mhat according to a binomial dist and then
## choose exactly mhat without replacement.  Faster than actually sampling when m << n^2.
function makeRandomMaskB( n::Int, m::Int )
	return drawWithoutReplacement(n , m)  # for now...
	p = m./n^2
	D = Binomial(n^2,p)
	mhat = rand(D)
	return drawWithoutReplacement( n, mhat )
end

# draws m things from [n]^2 without replacement.  I don't know how to do this well in Julia.
function drawWithoutReplacement( n::Int, m::Int )
	S = Set{Array{Int64, 1}}()
	while length(S) < m
		x = rand(1:n)
		y = rand(1:n)
		push!(S, [x,y] )
	end
	return [ x[1] for x in S],[ x[2] for x in S ]
end



# subsample a matrix A according to [xs],[ys].    If an entry appears twice,
# include it only once 
function subsample( A::Array, xs::Array, ys::Array )
	n = size(A,1)
	S = sparse(xs,ys, [ A[xs[j],ys[j]] for j=1:length(xs) ], n, n, max)	
	return S
end


# choose (about) m entries from A, and return a sparse matrix.
# This does NOT renormalize.
function subsample(A::Array, m::Int)
	n = size(A,1)
	xs,ys = makeRandomMask(n,m)
	S = subsample(A, xs,ys )
	return S
end

# subsample a matrix A = X*Y, where X is n by k and Y is k by n,
# according to [xs],[ys].  If an entry appears twice, include it only once.
function subsampleFactor( X::Array, Y::Array, xs::Array, ys::Array )
	n = size(X, 1)
	S = sparse( xs, ys, [ convert( Float64, (X[xs[j],:] * Y[:, ys[j]])[1,1]) for j=1:length(xs) ], n, n, max )
	return S
end

function subsampleFactor( X::Array, Y::Array, m::Int )
	n = size(X, 1)
	xs,ys = makeRandomMask(n,m)
	S = subsampleFactor( X, Y, xs, ys )
	return S
end

# Chop up an array lst into pieces of size sizes.
function chopUp( lst::Array{Int64, 1}, sizes::Array{Int64,1} )
	if sum(sizes) > length(lst)
		error("Cannot chop into such large pieces!", length(lst), " into ", lst2str(sizes))
	end
	shuffle!(lst)
	ret = Array(Array, length(sizes) )
	current = 1
	for i=1:length(sizes)
		ret[i] = lst[ current: current + sizes[i] - 1 ]
		current += sizes[i]
	end
	return ret
end

# Take a sparse matrix A and chop it into peices with sizes given by sizes.
function chopUp( A::SparseMatrixCSC, sizes::Array{Int} )
	n = size(A,1)
	xs,ys,vals = findnz( A )
	pieces = chopUp( collect(1:length(xs)) , sizes )
	xsnew =  [xs[i] for i in pieces[1] ] 
	ysnew =  [xs[i] for i in pieces[1] ]
	vsnew = [ vals[i] for i in pieces[1] ]
	return [ sparse( convert(Array{Int}, [xs[i] for i in piece]), convert( Array{Int}, [ys[i] for i in piece]), [vals[i] for i in piece], n, n, max ) for piece in pieces ]
end

function trunc(x, y, goToZero=false)
	if abs(x) < y
		return x
	else
		if goToZero
			return 0
		else
			return y*sign(x)
		end
	end
end

# no built-in map for sparse matrices??
function trunc(A::SparseMatrixCSC, y, goToZero=false)
	xs,ys,vs = findnz(A)
	for k=1:length(xs)
		A[xs[k],ys[k]] = trunc( vs[k] ,y , goToZero)
	end
end

function trunc(A::Array, y, goToZero=false)
	#B = map( (x)->trunc(x,y,goToZero), A)
	for i=1:size(A,1)
		for j=1:size(A,2)
			A[i,j] = trunc( A[i,j], y, goToZero )
		end
	end
end
	
# julia doesn't have a sparse SVD?  Really?
function svdPowerMethod(A, r, L)
	X = randn( size(A,1), r )
	for t=1:L
		Y = transpose(A) * (A * X )
		X,tmp = qr(Y)
	end
	sigmas = [ sqrt( norm( convert( Array{Float64}, transpose(A) * (A*X[:,i]) ) , 2) ) for i=1:r ]
	return X,sigmas
end

# svdPowerMethod for low-rank matrices, which are stored as an array of coefficients and vectors
function svdPowerMethod( coeffs, vecs, r, L )
	X = randn( length( vecs[1] ), r )
	for t=1:L
		Y = multBy( coeffs, vecs, X )
		X,tmp = qr(Y)
	end
	sigmas = [  norm( convert( Array{Float64},  multBy(coeffs, vecs, X[:,i]) ) , 2)  for i=1:r ]
	return X,sigmas
end

function multBy( coeffs, vecs, X )
	ret = zeros( size(X) )
	for i=1:length(coeffs)
		c = coeffs[i]
		v = vecs[i]
		ret += c * v * (transpose(v) * transpose(transpose(X)) )
	end
	return ret
end


function randomOrthonormal(n, k )
	X = randn( n, k )
	Y, tmp = qr(X)
	return Y
end

# returns a dict D so that D[i] contains a sparse vector representation of the ith column.
function colLists(A::SparseMatrixCSC)
	n = size(A,1)
	J,I,V = findnz(A)
	D = Dict()
	for k=1:length(I)	
		if haskey(D,I[k])
			(D[I[k]])[J[k]] = V[k]
		else
			D[I[k]] = sparsevec( [J[k]], [V[k]],n )
		end
	end
	return D
end


function randomLowRank(n,k, spectrum=-1)
	if spectrum == -1
		spectrum = [1.0 for i=1:k]
	end
	U = randn(n,k)
	Q,R = qr(U)
	return Q,spectrum #, Q * diagm(spectrum) * transpose(Q)
end

# U is a dense nxk matrix, returns (max row norm of U)/sqrt(k/n)
function coherence(U)
	return maximum( [ norm( U[i,:],2 ) for i=1:size(U,1) ] )/ sqrt(size(U,2)/size(U,1))
end

# X and U are n x k with orthogonal rows
function sinTheta( X, U )
	return norm( U - X*( transpose(X)*U ) )
end
function sinThetaFrob(X, U)
	return normfro(U - X*( transpose(X)*U ) )
end


############## thing to gather and plot data from deflate runs


# add some data to the tracker
function updateTracker( t::Tracker2, observations::Array{Float64,1}, typeString::ASCIIString )
	t.current += 1
	value = tuple( t.current, observations, typeString )
	push!(t.data,value)
end

# add the principal angles between U and X to the tracker
function updateTracker( t::Tracker2, X::Array{Float64,2}, U::Array{Float64,2} , typeString::ASCIIString)
	data = svd( U - X * (transpose(X) * U ))[2]	
	updateTracker(t, data, typeString)
end

function trackerLength( t::Tracker2 )
	return length(t.data)
end

#######

function lst2str( lst ) 
	str = ""
	for x in lst
		str = string(str, ",", string(x))
	end
	return str
end	


###

function entrywiseMedian( lst )
	return map( (x,y...)->median([x,y...]), lst...) 
end
		
### stuff for FW

function getEntry( i::Int,j::Int, coeffs::Array{Float64}{1}, vecs::Array{Array{Float64}}{1} ,tmax::Int)
	ret = 0
	for t=1:tmax
		ret += coeffs[t] * vecs[t][i] * vecs[t][j]	
	end
	return ret	
end
function getMat( coeffs::Array{Float64}{1}, vecs::Array{Array{Float64}}{1},tmax::Int )
	return sum( [ coeffs[i] * vecs[i] * transpose(vecs[i]) for i=1:tmax] )
end
## get frob( Q*Z - X*Y ) where Q,Z,X,Y are all skinny: Q,X are nxk, Z,Y are kxn
function factorFrobsq( Q, Z, X, Y )
	n = size(Q,1)
	# compute tr( (QZ - XY)(QZ -XY)^T )
	# first term: QZZ^TQ^T
	m1 = Z * transpose(Z) # kxk
	tr1 = sum( [ (Q[i,:] * m1 * transpose(Q)[:,i])[1,1] for i=1:n ] )
	# second term: QZY^TX^T 
	m2 = Z*transpose(Y)
	tr2 = sum( [(Q[i,:] * m2 * transpose(X)[:,i])[1,1] for i=1:n ] )
	# second.5 term: XYZ^TQ^T  (okay, I guess this is the same as m2) 
	#m25 = Y*transpose(Z)
	#tr25 = sum( [(X[i,:] * m25 * transpose(Q)[:,i])[1,1] for i=1:n ] )
	# third term: XYY^TX^T
	m3 = Y * transpose(Y) # kxk
	tr3 = sum( [ (X[i,:] * m3 * transpose(X)[:,i])[1,1] for i=1:n ] )
	return tr1 - 2*tr2  + tr3
end

function factorFrob(Q,Z,X,Y)	
	return convert( Float64, sqrt( factorFrobsq(Q,Z,X,Y) ))
end


