
####################################
#
#
#  Some tests for altmin.jl
#
#
####################################

using(Winston)
using(Color)
include("altmin.jl")
include("frankWolfe.jl")

################################################################################
#
# getData
#
# runs a trialfn (basicAM, initAM, deflateAM, etc) on an
# nxn symmetric matrix with the specified spectrum and random
# eigenvectors.  
# ms = Array of number of measurements to try
# trials = number of times to do each measurement level
# args... = additional args for trialfn
# 
# returns three arrays: 
#    an array of the number of measurements.
#    an array of the errors
#    an array of Tracker2 objects for drawing pretty pictures. 
##########################################################################################################
function getData( n, ms, spectrum, trials, trialfn, args...; debug=false )
	data = Array( Float64,  trials*length(ms) )
	msactual = Array( Float64, trials*length(ms) )
	trackers = Array( Tracker2, trials*length(ms) )
	count = 1
	for i in 1:length(ms)
		for t=1:trials
			err, mactual, tracker = trialfn(n, spectrum, ms[i], args,debug=debug)
			data[count] = err
			msactual[count]= mactual
			trackers[count] = tracker	
			count += 1
		end
	end
	return msactual, data, trackers
end

# takes a list of ms, errs as returned by getData,
# and writes the data to a file with filename fname
# starts the file with tags.
function writeData( ms, errs, tags, fname )
	F = open( fname, "w")
	write(F, tags)
	write(F, "\n")
	for i in 1:length(ms)
		write(F, string(ms[i]))
		write(F, " ")
		write(F, string(errs[i]))
		write(F, "\n")
	end
	close(F)
end

# plots (ms, errs) on plot p, with legend given by tags and color col.
function plotData(ms, errs, p, tags, col,pos=.9)
	c = Points( ms, errs ,color=col)
	setattr(c, label=string(tags))
	style(c, kind="filled circle")
	l = Legend(.6,pos,{c})
	add(p,c,l)
end



# plots data from Tracker2 objects which are created from AM and Deflate runs
# trackers is an Array of Tracker2's.
# col is an RGB color, like RGB(1,0,0)
# if the "write" flag is on, it will also write the tracker's data to files with the tag "ftag"
function plotTracker( trackers , k::Int, p::FramedPlot , tags, col, pos=.9; writeflag=false, ftag="")
	N = maximum( [ trackerLength(t) for t in trackers ] )  # the should all have the same length anyway
	T = length(trackers)
	ms = Dict()
	ms["AM"] = Dict()
	ms["Deflate"] = Dict()
	if writeflag
		try 
			mkdir( string(ftag, "_tracking" ))
		catch
			println("Directory ", ftag,"_tracking already exists, overwriting contents...")
		end
	end
	for i=1:k
		if writeflag 
			F = open( string( ftag, "_tracking/", ftag, "_track_", i, ".dat" ), "w" )
		end
		ms["AM"][i] = Dict{Int64, Float64}()  # maps ints to floats (error)( zeros(Int64, N), zeros( Float64, N) )
		ms["Deflate"][i] = Dict{Int64, Float64}()
		# sort the principal angles from 1 to k, and fill in any missing biggest with "1"
		for t in trackers
			for x in t.data  # x looks like ( i, [p1,p2], type )
				step = x[1]
				angs = sort(x[2])
				tag = x[3] # either "AM" or "Deflate" -- different types of steps
				if i > length(angs)
					val = 1.0/T
				else
					val = angs[i]/T
				end
				if haskey( ms[tag][i], step ) 
					ms[tag][i][step] += val
				else
					ms[tag][i][step] = val
				end
			end
		end
		# if I were a good person I would have done this all in an array to begin with.
		amkeys = collect( keys( ms["AM"][i] ) ) 
		defkeys = collect( keys( ms["Deflate"][i] ) )
		amvals = collect( values( ms["AM"][i] ) )
		defvals = collect( values( ms["Deflate"][i] ) )
		if writeflag
			for j=1:length(amkeys)
				write( F, string( amkeys[j], "\t", amvals[j], "\t AM \n" ))
			end
			for j=1:length(defkeys)
				write( F, string( defkeys[j], "\t", defvals[j], "\t Deflate \n" ))
			end
			close(F)
		end
		AMpts = Points( amkeys, amvals, color=uCol(col,i/k) )
		Deflpts = Points( defkeys, defvals, color=uCol(col,i/k) )
		allOfThem = zeros( Float64, length(amkeys) + length(defkeys) )
		for z=1:length(allOfThem)
			if haskey( ms["AM"][i] , z )
				allOfThem[z] = ms["AM"][i][z]
			else
				allOfThem[z] = ms["Deflate"][i][z]
			end
		end
		C = Curve( 1:length(allOfThem), allOfThem, color = uCol(col, i/k))
		style(AMpts, kind="filled circle")
		style(Deflpts, kind="circle")
		if i==k
			setattr(C, label=tags)
			l = Legend(.1, pos, {C})
			add(p,C,l)
		else
			add(p,C)
		end
		add(p,AMpts)
		add(p, Deflpts) 
	end
end

function uCol( col, frac1 )
	frac = (frac1*5 + 1)/6
	return RGB( col.r * frac, col.b*frac, col.g*frac )
end



########
#
# generateData: makes matrices with a specified spectrum and random eigenvectors, and runs three algorithms on them: basic SVD, FW, and Deflate.
#
# Writes the data to a file and produces a plot.  There's a python script in this directory which makes a prettier plot from the files.
#
##########
# n is the dimension of the matrix.
# ms is a list of the total number of measurements to do.
# spectrum is the spectrum to use.
# T is the number of anticipated blocks in the spectrum (the algorithm does not use this; it just helps for getting the number of measurements on target).
# M is the number of measurements for SoftDeflate to use per iteration in each run of AM/SVD.  
#      So T(L+1)M + M = total measurements: M for the initial SVD, and then for T epochs, M measurements for each of L iterations of AM, 
#      and M more for the truncation/SVD step.  Right now the algorithm does not do medians or smoothing.
# tag is a filename header 
# trials is the number of trials to do.
##########
function generateData( n, ms , spectrum, T, M, tag, trials=10)
	tags = string("n: ", n, ", Trials: ", trials, ", Spectrum: ", lst2str(spectrum))
	p = FramedPlot(title=tags, xlabel="m", ylabel="error (Frobenius)", ylog=true)
	# get Deflation data: this uses M samples per AM iteration, and goes up to m.
	println("Deflate...")
	msDeflate, dataDeflate, trackerDeflate = getData( n, ms, spectrum, trials, doTrialDeflate, M, M, T,debug=false)
	writeData( msDeflate, dataDeflate, tags, string(tag, "_Deflate_", T, ".out") )
	plotData(msDeflate, dataDeflate, p, string("SoftDeflate"), "green", .75)
	#get SVD data
	println("SVD...")
	msSVD, dataSVD, tracker = getData( n, ms, spectrum, trials, doTrialSVD, -1,debug=false)
	writeData( msSVD, dataSVD, tags, string(tag, "_SVD.out") )
	plotData(msSVD, dataSVD, p, string("SVD"), "blue", .8)
	# get FW data, for two different values.
	epss = [.05]
	cols = ["red"]
	poss = [.85]
	for i=1:1
		eps = epss[i]
		println("FW (", eps, ")...")
		msFW, dataFW, tracker = getData( n, ms, spectrum, trials, doTrialFW, eps)
		writeData(msFW, dataFW, string(tags, ", eps:",eps), string(tag, "_FW_",eps,".out") )
		plotData(msFW, dataFW, p, string("FW:",eps), cols[i], poss[i])
	end

	return p
end

######################################################################
#
# How well does FW converge, on the samples and in general?
#
#####################################################################
function trackFW( n, m, spectrum, eps, tag )
	trials = 1
	msFW, dataFW, trackerFW = getData( n, [m], spectrum, trials, doTrialFW, eps ,debug=true)
	# now make sense of the data in the tracker.  trackerFW[1].data is a Set of tuples ( step, [ error_on_samples, total_frob_error ], "placeholderstring" )
	t = trackerFW[1]
	steps = 1:t.current
	onsampleerrs = zeros( Float64, t.current )
	totalerrs = zeros( Float64, t.current )
	for x in t.data
		step = x[1]
		onsampleerr = x[2][1]
		totalerr = x[2][2]
		onsampleerrs[step] = onsampleerr
		totalerrs[step] = totalerr
	end
	p = FramedPlot(title=string( "Performance of FW: spectrum: ", lst2str(spectrum), " n=",n," m=", m ), xlabel="iteration", ylabel="error (frobenius norm)", ylog=true)	
	#plotData( steps, onsampleerrs, p, "On the samples", "red", .9 )
	#plotData( steps, totalerrs, p, "Total error", "blue", .8 )
	writeData( steps, onsampleerrs, string( "Error on samples for FW; spectrum=", lst2str(spectrum), " m=", m, " n=", n ), string( tag, "_onsamples.out" ) )
	writeData( steps, totalerrs, string( "Frobenius error for FW; spectrum=", lst2str(spectrum), " m=", m, " n=", n ), string( tag, "_froberr.out" ) )
	return p
end

# run the above function a bunch of times, for both a flat and not-flat spectrum.
function trackFWalot( n, m, tag , trials)
	for t =1:trials
		trackFW( n, m, [1,1,1], 1/30, string(tag, "_flat_", t ) )
		trackFW( n, m, [1,1,.1], 1/30, string(tag, "_spiky_", t ) )
	end
end

#####################################################################
#		
#  How does AM compare to our soft-deflation-AM?
#
#  This does one run of each and plots the progress of all of the 
#  principal angles.
#
#####################
#
# n is the dimension of the matrix.
# m is the number of measurements total.
# M is the number of measurements per iteration of AM / Deflation
function testAMvsDeflateAM( n, L, M, ftag )
	trials = 1
	m = L*M
	gp = 2  # this is how many gaps we expect...otherwise softDeflate uses way too few observations and it's not fair.
	spectrum = [1,1,1,.01,.01,.01]
	k = length(spectrum)
	tags = string( "n=",string(n), " Meas. per step=", M)
	p = FramedPlot(title=tags, xlabel="steps", ylabel="sin(Theta(X,U))", ylog=true)
	println("================================================")
	println("Starting BasicAM...m=",m, " L=",L," miter=",M)
	println("================================================")
	ms1, dataSVD,trackerSVD = getData( n, [m], spectrum, trials, doTrialBasicAM, M,-1,debug=true)
	println("================================================")
	println("Done with BasicAM...starting deflationAM")
	println("================================================")
	ms2, datadeflate,trackerdeflate = getData(n, [m], spectrum, trials, doTrialDeflate, M, M, gp,debug=true)
	plotTracker( trackerSVD, k, p, "AM", RGB(1,0,0) ,.1 , writeflag=true, ftag=string(ftag, "_AM"))
	plotTracker( trackerdeflate, k, p, "Def", RGB(0,1,0), .2 , writeflag=true, ftag=string( ftag, "_Deflate"))
	return p, ms1, ms2, dataSVD, datadeflate
end


########################################################
#
#   Does median actually help in AM?
#
#############
#
#
function testMedian(n, ms, miter, trials, tag)
	spectrum = [1,1,1]
	k=length(spectrum)
	tags = string( "n=",string(n), " spectrum:", lst2str(spectrum), " trials:", trials, " mperiter:", miter )
	p = FramedPlot(title=tags, xlabel="observations", ylabel="error (Frobenius)", ylog=true)
	println("------- NO MEDIAN -----------")
	ms1, data1,tracker1 = getData( n, ms, spectrum, trials, doTrialBasicAM, miter, -1, debug=false)
	println("----------MEDIAN-------------")
	ms2, data2,tracker2 = getData( n, ms, spectrum, trials, doTrialBasicAM, miter, 3, debug=false)
	plotData( ms1, data1, p, "No median", "red", .9 )
	plotData( ms2, data2, p, "Median", "blue", .8 )
	writeData( ms1, data1, string( "testMedian:", tags, ": no median" ), string( tag, "_nomedian.out" ) )
	writeData( ms2, data2, string( "testMedian:", tags, ": median" ), string( tag, "_median.out" ) )
	return p
end


### Test the effect of initialization on AM.
function testInit( n, spectrum, L, M, ftag )
	trials = 1
	m = L*M
	k = length(spectrum)
	tags = string( "n=",string(n), " Meas. per step=", M)
	p = FramedPlot(title=tags, xlabel="steps", ylabel="sin(Theta(X,U))", ylog=true)
	println("================================================")
	println("Starting BasicAM...m=",m, " L=",L," miter=",M)
	println("================================================")
	ms1, dataSVD,trackerSVD = getData( n, [m], spectrum, trials, doTrialBasicAM, M,-1,debug=true)
	println("================================================")
	println("Starting BasicAM...m=",m, " L=",L," miter=",M)
	println("================================================")
	ms2, data2,tracker2 = getData( n, [m], spectrum, trials, doTrialInitAM, M,M,-1,debug=true)
	plotTracker( trackerSVD, k, p, "Random Init", RGB(1,0,0) ,.1 , writeflag=true, ftag=string(ftag, "_Rand"))
	plotTracker( tracker2, k, p, "SVD Init", RGB(0,1,0), .2 , writeflag=true, ftag=string( ftag, "_Init"))
	return p
end


########### doTrial functions: ############################
# these take n, spectrum, m, and some other parameters, 
# and try out a recovery algorithm on a symmetric matrix 
# with the specified spectrum and random eigenvectors.
#
# error is measured according to getError, which right now is
# sin(theta) between correct and returned eigenspaces.
##############################################################

# alternating minimization, with truncated SVD for initialization.	
# args = (mstart, miter), the number of measurements to use for the initial SVD and for each iteration.
# m = total number of observations
function doTrialInitAM(n, spectrum, m, args;debug=false)
	mstart = args[1]
	miter = args[2]
	k = length(spectrum)
	Q, S = randomLowRank(n,k,spectrum)
	tracker = Tracker2( 0, Set() )
	X,Y,nsamples = initializedAM( subsampleFactor( Q, diagm(spectrum)*transpose(Q) ,m), mstart, miter, k, debug=debug, correctX=Q, tracker=tracker)	
	err = getError( X, Y, Q ,S)
	println("m:", nsamples, "\tERROR:", err)
	return err, nsamples, tracker
end	

# alternating minimization, with truncated SVD for initialization.	
# args = (miter,median)
# m = total number of observations
function doTrialBasicAM(n, spectrum, m, args; debug=false)
	miter = args[1]
	domedian = args[2]
	k = length(spectrum)
	Q, S = randomLowRank(n,k,spectrum)
	tracker = Tracker2( 0, Set() )
	X,Y,nsamples = basicAM( subsampleFactor( Q, diagm(spectrum)*transpose(Q) ,m), miter, k,debug=debug, correctX=Q, tracker=tracker, domedian=domedian)	
	err= getError( X, Y, Q ,S)
	println("m:", nsamples, "\tERROR:", err)
	return err,nsamples, tracker
end	

# just take the SVD of the sparsified matrix; args is ignored
function doTrialSVD(n, spectrum, m, args;debug=false)
	k = length(spectrum)
	Q, S = randomLowRank(n,k,spectrum)
	B = subsampleFactor(Q, diagm(S)*transpose(Q) ,m)
	nsamples = nnz(B)
	X,sigmas = svdPowerMethod(B, k, 100)
	tracker = Tracker2( 0, Set() ) # no interesting data to track.
	err = getError( X, diagm(convert( Array{Float64,1}, sigmas))*transpose(X), Q ,S)
	println("m:", nsamples, "\tERROR:", err)
	return err,nsamples,tracker
end	

## do Frank-Wolfe ("Hazan's algorithm" in [Jaggi-Sulovsky'2010]), for symmetric matrices
## args is epsilon, the accuracy parameter.
function doTrialFW(n, spectrum, m, args;debug=false)
	eps = args[1]
	tracker = Tracker2( 0, Set() )
	k = length(spectrum)
	Q,S = randomLowRank(n,k,spectrum)
	tr = sum(S) 
	if debug
		normalA = Q * diagm(S) * transpose(Q)/tr # now A has trace 1.
	end

	# do FW
	if debug
		coeffs, vecs, nsamples = FW( subsampleFactor(Q, diagm(S) * transpose(Q)/tr, m ) , eps, debug=debug, correctQ=Q,correctS = S/tr, tracker=tracker)
	else
		coeffs, vecs, nsamples = FW( subsampleFactor(Q, diagm(S) * transpose(Q)/tr, m ) , eps, debug=debug, correctQ=-1, correctS=-1,tracker=tracker)
	end

	# scale back up the coefficients and find the error.
	X, sigmas = svdPowerMethod( tr* coeffs, vecs, k, 100 )
	err = sinTheta( X, Q) 
	
	#V = hcat( vecs... )  # this is a tall skinny matrix
	#D = tr*diagm( coeffs ) 
	#err2 = factorFrob( V, D*transpose(V), Q, diagm(S)*transpose(Q) )  # get frobenius error 

	#	err2tmp = normfro( getMat( coeffs*tr, vecs, length(coeffs) ) - Q * diagm(S) * transpose(Q) )    # not the smart thing to do.
	#println("ERROR:\t",err) #, "\t", err2)
	println("m:", nsamples, "\tERROR:", err )
	return err, nsamples, tracker
end
# our deflation-based AM algorithm.  m is the TOTAL number of samples.
# params = [mDeflate, mAM, T], as in the definition of DeflateAM
function doTrialDeflate( n, spectrum, m, params ;debug=false)
	mDeflate = params[1]
	mAM = params[2]
	T = params[3]
	k = length(spectrum)
	Q, S = randomLowRank(n,k,spectrum)
	tracker = Tracker2( 0, Set() )
	# mu0 = 10*(k^(3/2)*mustar + log(n)) not using this now....
	X,Y,nsamples = deflationAM( subsampleFactor(Q, diagm(spectrum)*transpose(Q), m), mDeflate, mAM, k, T=T, debug=debug, correctX=Q, correctSigmas=S,tracker=tracker) 	
	err= getError( X, Y,  Q, S )
	println("m:", nsamples, "\tERROR:", err)
	return err,nsamples, tracker
end

	
# Choose our favorite error function.
function getError( X, Y, Q , S)
	### sin(theta):
	#sinError = sinTheta( X, Q )
	### || sin(Theta) ||_2:
	# froSinError = sinThetaFrob(X, Q)
	## reconstruction error in frobenius norm:
	froError = normfro( Q*diagm(S)*transpose(Q) - X*Y )
	return froError
	#return factorFrob( X, Y, Q, diagm(S) * transpose(Q) )
	#return sinError
end


################################################################################
#
#  Debugging tests to make sure I've implemented this like a sane person.
#	
######################################################
#

function testScaling()
	spectrum = [1,1,1]
	basen = 1000
	basem = 500000
	baseT = 50000

	factors = [1,1.25,1.5,1.75,2,2.25,2.5,2.75,3]
	datas = [0.0 for i=1:length(factors)]
	for i=1:length(factors)
		mul = factors[i]
		n= convert(Int, round(mul*basen) )
		m = convert(Int, round(mul*basem) )
		T = convert(Int, round(mul*baseT) )
		println(mul,":\t n=", n, " m=", m, " T=", T )
		tic()
		doTrialBasicAM( n, spectrum, m, (T, -1), debug=false )
		tm = toq()
		println("\t TIME:", tm)
		datas[i] = tm
	end
	return datas
end	
